@@ -34,14 +34,17 @@
 #include <unistd.h>
 #include <sys/stat.h>
 #include "yp.h"
-
+#include <sys/socket.h>
+#include <arpa/inet.h>
+#include <netdb.h>
 
 #ifndef _PATH_YP
 #define _PATH_YP   "/var/yp"
 #endif
 char *path_ypdb = _PATH_YP;
 
-int debug_flag = 1;
+int debug_flag = 0;
+int dns_flag   = 0;
 
 
 static void Gperror(const char *str)
@@ -50,18 +53,37 @@ static void Gperror(const char *str)
 }
 
 
+/*
+** Return 1 if request comes from an authorized host
+**
+** XXX This function should implement the "securenets" functionality
+*/
+static int is_valid_host(struct sockaddr_in *sin)
+{
+    return 1;
+}
+
+
 void *ypproc_null_2_svc(void *dummy,
 			struct svc_req *rqstp)
 {
     static int foo;
-
+    struct sockaddr_in *rqhost;
     
+
+    rqhost = svc_getcaller(rqstp->rq_xprt);
+    if (!is_valid_host(rqhost))
+	return NULL;
+
     if (debug_flag)
-	fprintf(stderr, "ypproc_null()\n");
+	fprintf(stderr, "ypproc_null() [From: %s:%d]\n",
+		inet_ntoa(rqhost->sin_addr),
+		ntohs(rqhost->sin_port));
 
     return (void *) &foo;
 }
 
+    
 /*
 ** Return 1 if the name is a valid domain name served by us, else 0.
 */
@@ -85,14 +107,23 @@ static int is_valid_domain(const char *domain)
 }
 
 
+
 bool_t *ypproc_domain_2_svc(domainname *name,
 			    struct svc_req *rqstp)
 {
     static bool_t result;
-
-
+    struct sockaddr_in *rqhost;
+    
+    
+    rqhost = svc_getcaller(rqstp->rq_xprt);
+    if (!is_valid_host(rqhost))
+	return NULL;
+	
     if (debug_flag)
-	fprintf(stderr, "ypproc_domain(\"%s\")\n", *name);
+	fprintf(stderr, "ypproc_domain(\"%s\") [From: %s:%d]\n",
+		*name,
+		inet_ntoa(rqhost->sin_addr),
+		ntohs(rqhost->sin_port));
 
     if (is_valid_domain(*name))
 	result = TRUE;
@@ -111,9 +142,18 @@ bool_t *ypproc_domain_nonack_2_svc(domainname *name,
 				   struct svc_req *rqstp)
 {
     static bool_t result;
+    struct sockaddr_in *rqhost;
+
+    
+    rqhost = svc_getcaller(rqstp->rq_xprt);
+    if (!is_valid_host(rqhost))
+	return NULL;
 
     if (debug_flag)
-	fprintf(stderr, "ypproc_domain_nonack(\"%s\")\n", *name);
+	fprintf(stderr, "ypproc_domain_nonack(\"%s\") [From: %s:%d]\n",
+		*name,
+		inet_ntoa(rqhost->sin_addr),
+		ntohs(rqhost->sin_port));
     
     if (!is_valid_domain(*name))
     {
@@ -163,6 +203,9 @@ static GDBM_FILE open_database(const char *domain,
 }
 
 
+#define F_ALL   0x01
+#define F_NEXT  0x02
+
 /*
 ** Get a record from a GDBM database.
 */
@@ -170,7 +213,7 @@ int read_database(GDBM_FILE dbp,
 		  const datum *ikey,
 		  datum *okey,
 		  datum *dval,
-		  int all_flag)
+		  int flags)
 {
     int first_flag = 0;
     datum nkey, ckey;
@@ -182,10 +225,17 @@ int read_database(GDBM_FILE dbp,
 	first_flag = 1;
     }
     else
-	ckey = *ikey;
+    {
+	if ((flags & F_NEXT))
+	    ckey = gdbm_nextkey(dbp, *ikey);
+	else
+	    ckey = *ikey;
+    }
     
     if (ckey.dptr == NULL)
+    {
 	return YP_NOKEY;
+    }
 
     while (1)
     {
@@ -197,7 +247,9 @@ int read_database(GDBM_FILE dbp,
 		free(ckey.dptr);
 
 	    if (ikey && ikey->dptr != NULL)
+	    {
 		return YP_NOKEY;
+	    }
 	    else
 		if (first_flag)
 		    return YP_BADDB;
@@ -205,7 +257,7 @@ int read_database(GDBM_FILE dbp,
 		    return YP_FALSE;
 	}
 	
-	if (all_flag || strncmp(dval->dptr, "YP_", 3) != 0)
+	if ((flags & F_ALL) || strncmp(dval->dptr, "YP_", 3) != 0)
 	{
 	    if (okey)
 		*okey = ckey;
@@ -256,10 +308,19 @@ ypresp_val *ypproc_match_2_svc(ypreq_key *key,
 			       struct svc_req *rqstp)
 {
     static ypresp_val result;
+    struct sockaddr_in *rqhost;
+
+    
+    rqhost = svc_getcaller(rqstp->rq_xprt);
+    if (!is_valid_host(rqhost))
+	return NULL;
 
     if (debug_flag)
     {
-	fprintf(stderr, "ypproc_match():\n");
+	fprintf(stderr, "ypproc_match(): [From: %s:%d]\n",
+		inet_ntoa(rqhost->sin_addr),
+		ntohs(rqhost->sin_port));
+		
 	fprintf(stderr, "\t\tdomainname = \"%s\"\n",
 		key->domain);
 	fprintf(stderr, "\t\tmapname = \"%s\"\n",
@@ -292,7 +353,7 @@ ypresp_val *ypproc_match_2_svc(ypreq_key *key,
 	    qdat.dsize = key->key.keydat_len;
 	    qdat.dptr = key->key.keydat_val;
 
-	    result.stat = read_database(dbp, &qdat, NULL, &rdat, 1);
+	    result.stat = read_database(dbp, &qdat, NULL, &rdat, F_ALL);
 
 	    if (result.stat == YP_TRUE)
 	    {
@@ -303,9 +364,9 @@ ypresp_val *ypproc_match_2_svc(ypreq_key *key,
 	    gdbm_close(dbp);
 	}
     }
-
-    if (debug_flag)
-    {
+	
+	if (debug_flag)
+	{
 	if (result.stat == YP_TRUE)
 	    fprintf(stderr, "\t-> Value = \"%.*s\"\n",
 		    (int) result.val.valdat_len,
@@ -313,6 +374,72 @@ ypresp_val *ypproc_match_2_svc(ypreq_key *key,
 	else
 	    fprintf(stderr, "\t-> Error #%d\n", result.stat);
     }
+
+
+    /*
+    ** Do the jive thing if we didn't find the host in the YP map
+    ** and we have enabled the magic DNS lookup stuff.
+    **
+    ** XXX Perhaps this should be done in a sub-process for performance
+    **     reasons. Later.
+    */
+    if (result.stat != YP_TRUE && dns_flag)
+    {
+	struct hostent *he;
+	key->key.keydat_val[key->key.keydat_len] = '\0';
+
+
+	if (debug_flag)
+	    fprintf(stderr, "Doing DNS lookup of %s\n", key->key.keydat_val);
+	
+	if (strcmp(key->map, "hosts.byname") == 0)
+	    he = gethostbyname(key->key.keydat_val);
+	else  if (strcmp(key->map, "hosts.byaddr") == 0)
+	    he = gethostbyaddr(inet_addr(key->key.keydat_val), sizeof(long),
+			       AF_INET);
+	else
+	    he = NULL;
+	    
+	if (he)
+	{
+	    int len, i;
+	    char *cp;
+	    
+	    
+	    if (debug_flag)
+		fprintf(stderr, "\t-> OK (%s/%s)\n",
+			he->h_name,
+			inet_ntoa(he->h_addr));
+
+	    len = 16; /* Numerical IP address */
+	    len += strlen(he->h_name) + 1;
+	    for (i = 0; he->h_aliases[i]; i++)
+		len += strlen(he->h_aliases[i]) + 1;
+	    
+
+	    cp = malloc(len+1);
+	    strcpy(cp, inet_ntoa(he->h_addr));
+	    strcat(cp, " ");
+	    strcat(cp, he->h_name);
+
+	    for (i = 0; he->h_aliases[i]; i++)
+	    {
+		strcat(cp, " ");
+		strcat(cp, he->h_aliases[i]);
+	    }
+
+	    result.val.valdat_len = len;
+	    result.val.valdat_val = cp;
+	    result.stat = YP_TRUE;
+	}
+	else
+	{
+	    if (debug_flag)
+		fprintf(stderr, "\t-> Not Found\n");
+
+	    result.stat = YP_NOKEY;
+	}
+    }
     
     return &result;
 }
@@ -323,11 +450,19 @@ ypresp_key_val *ypproc_first_2_svc(ypreq_key *key,
 				   struct svc_req *rqstp)
 {
     static ypresp_key_val result;
+    struct sockaddr_in *rqhost;
+    
 
+    rqhost = svc_getcaller(rqstp->rq_xprt);
+    if (!is_valid_host(rqhost))
+	return NULL;
 
     if (debug_flag)
     {
-	fprintf(stderr, "ypproc_first():\n");
+	fprintf(stderr, "ypproc_first(): [From: %s:%d]\n",
+		inet_ntoa(rqhost->sin_addr),
+		ntohs(rqhost->sin_port));
+		
 	fprintf(stderr, "\tdomainname = \"%s\"\n", key->domain);
 	fprintf(stderr, "\tmapname = \"%s\"\n", key->map);
 	fprintf(stderr, "\tkeydat = \"%.*s\"\n",
@@ -398,11 +533,19 @@ ypresp_key_val *ypproc_next_2_svc(ypreq_key *key,
 				  struct svc_req *rqstp)
 {
     static ypresp_key_val result;
-
+    struct sockaddr_in *rqhost;
     
+    
+    rqhost = svc_getcaller(rqstp->rq_xprt);
+    if (!is_valid_host(rqhost))
+	return NULL;
+
     if (debug_flag)
     {
-	fprintf(stderr, "ypproc_match():\n");
+	fprintf(stderr, "ypproc_match(): [From: %s:%d]\n",
+		inet_ntoa(rqhost->sin_addr),
+		ntohs(rqhost->sin_port));
+		
 	fprintf(stderr, "\tdomainname = \"%s\"\n", key->domain);
 	fprintf(stderr, "\tmapname = \"%s\"\n", key->map);
 	fprintf(stderr, "\tkeydat = \"%.*s\"\n",
@@ -454,7 +597,7 @@ ypresp_key_val *ypproc_next_2_svc(ypreq_key *key,
 	    gdbm_close(dbp);
 	}
     }
-
+    
     if (debug_flag)
     {
 	if (result.stat == YP_TRUE)
@@ -466,10 +609,12 @@ ypresp_key_val *ypproc_next_2_svc(ypreq_key *key,
 	else
 	    fprintf(stderr, "\t-> Error #%d\n", result.stat);
     }
-
+    
     return &result;
 }
 
+
+
 static void print_ypmap_parms(const struct ypmap_parms *pp)
 {
     fprintf(stderr, "\t\tdomain   = \"%s\"\n", pp->domain);
@@ -486,10 +631,19 @@ ypresp_xfr *ypproc_xfr_2_svc(ypreq_xfr *xfr,
 			     struct svc_req *rqstp)
 {
     static ypresp_xfr result;
+    struct sockaddr_in *rqhost;
+
+
+    rqhost = svc_getcaller(rqstp->rq_xprt);
+    if (!is_valid_host(rqhost))
+	return NULL;
 
     if (debug_flag)
     {
-	fprintf(stderr, "ypproc_xfr_2_svc():\n\tmap_parms:\n");
+	fprintf(stderr, "ypproc_xfr_2_svc(): [From: %s:%d]\n\tmap_parms:\n",
+		inet_ntoa(rqhost->sin_addr),
+		ntohs(rqhost->sin_port));
+	
 	print_ypmap_parms(&xfr->map_parms);
 	fprintf(stderr, "\t\ttransid = %u\n", xfr->transid);
 	fprintf(stderr, "\t\tprog = %u\n", xfr->prog);
@@ -508,9 +662,18 @@ void *ypproc_clear_2_svc(void *dummy,
 			 struct svc_req *rqstp)
 {
     static int foo;
+    struct sockaddr_in *rqhost;
+    
+
+    rqhost = svc_getcaller(rqstp->rq_xprt);
+    if (!is_valid_host(rqhost))
+	return NULL;
 
     if (debug_flag)
-    	fprintf(stderr, "ypproc_clear_2_svc()\n");
+    	fprintf(stderr, "ypproc_clear_2_svc() [From: %s:%d]\n",
+		inet_ntoa(rqhost->sin_addr),
+		ntohs(rqhost->sin_port));
+		
 
     return (void *) &foo;
 }
@@ -518,13 +681,10 @@ void *ypproc_clear_2_svc(void *dummy,
 
 static int ypall_close(void *data)
 {
-    if (debug_flag)
-	if (data == NULL)
-	    fprintf(stderr, "ypall_close() called with NULL pointer.\n");
-	else
-	    fprintf(stderr, "ypall_close() called.\n");
+    if (debug_flag && data == NULL)
+	fprintf(stderr, "ypall_close() called with NULL pointer.\n");
 	    
-    gdbm_close(*(GDBM_FILE *) data);
+    gdbm_close((GDBM_FILE) data);
     return 0;
 }
 
@@ -533,19 +693,14 @@ static int ypall_encode(ypresp_key_val *val,
 			void *data)
 {
     datum dkey, dval, okey;
-    int status;
 
-    if (debug_flag)
-	fprintf(stderr, "ypall_encode() called with key \"%.*s\".\n",
-		(int) val->key.keydat_len,
-		val->key.keydat_val);
     
     dkey.dptr = val->key.keydat_val;
     dkey.dsize = val->key.keydat_len;
     
-    status = read_database(*(GDBM_FILE *) data, &dkey, &okey, &dval, 0);
+    val->stat = read_database((GDBM_FILE) data, &dkey, &okey, &dval, F_NEXT);
 
-    if (status == YP_TRUE)
+    if (val->stat == YP_TRUE)
     {
 	val->key.keydat_val = okey.dptr;
 	val->key.keydat_len = okey.dsize;
@@ -553,8 +708,9 @@ static int ypall_encode(ypresp_key_val *val,
 	val->val.valdat_val = dval.dptr;
 	val->val.valdat_len = dval.dsize;
     }
+
     
-    return status;
+    return val->stat;
 }
 
 
@@ -563,11 +719,19 @@ ypresp_all *ypproc_all_2_svc(ypreq_nokey *nokey,
 {
     static ypresp_all result;
     extern xdr_ypall_cb_t xdr_ypall_cb;
+    struct sockaddr_in *rqhost;
 
     
+    rqhost = svc_getcaller(rqstp->rq_xprt);
+    if (!is_valid_host(rqhost))
+	return NULL;
+
     if (debug_flag)
     {
-	fprintf(stderr, "ypproc_all_2_svc():\n");
+	fprintf(stderr, "ypproc_all_2_svc(): [From: %s:%d]\n",
+		inet_ntoa(rqhost->sin_addr),
+		ntohs(rqhost->sin_port));
+		
 	fprintf(stderr, "\t\tdomain = \"%s\"\n", nokey->domain);
 	fprintf(stderr, "\t\tmap = \"%s\"\n", nokey->map);
     }
@@ -607,17 +771,11 @@ ypresp_all *ypproc_all_2_svc(ypreq_nokey *nokey,
 
 		xdr_ypall_cb.u.encode = ypall_encode;
 		xdr_ypall_cb.u.close  = ypall_close;
-		xdr_ypall_cb.data = (void *) &dbp;
+		xdr_ypall_cb.data = (void *) dbp;
 
-		if (debug_flag)
-		    fprintf(stderr, "Returning because YP_TRUE\n");
-		
 		return &result;
 	    }
 
-	    if (debug_flag)
-		fprintf(stderr, "Closing ypall GDBM file.\n");
-	    
 	    gdbm_close(dbp);
 	}
     }
@@ -630,11 +788,19 @@ ypresp_master *ypproc_master_2_svc(ypreq_nokey *nokey,
 				   struct svc_req *rqstp)
 {
     static ypresp_master result;
-
+    struct sockaddr_in *rqhost;
     
+    
+    rqhost = svc_getcaller(rqstp->rq_xprt);
+    if (!is_valid_host(rqhost))
+	return NULL;
+
     if (debug_flag)
     {
-	fprintf(stderr, "ypproc_master_2_svc():\n");
+	fprintf(stderr, "ypproc_master_2_svc(): [From: %s:%d]\n",
+		inet_ntoa(rqhost->sin_addr),
+		ntohs(rqhost->sin_port));
+		
 	fprintf(stderr, "\t\tdomain = \"%s\"\n", nokey->domain);
 	fprintf(stderr, "\t\tmap = \"%s\"\n", nokey->map);
     }
@@ -692,10 +858,19 @@ ypresp_order *ypproc_order_2_svc(ypreq_nokey *nokey,
 				 struct svc_req *rqstp)
 {
     static ypresp_order result;
+    struct sockaddr_in *rqhost;
+    
+    
+    rqhost = svc_getcaller(rqstp->rq_xprt);
+    if (!is_valid_host(rqhost))
+	return NULL;
 
     if (debug_flag)
     {
-	fprintf(stderr, "ypproc_order_2_svc():\n");
+	fprintf(stderr, "ypproc_order_2_svc(): [From: %s:%d]\n",
+		inet_ntoa(rqhost->sin_addr),
+		ntohs(rqhost->sin_port));
+		
 	fprintf(stderr, "\t\tdomain = \"%s\"\n", nokey->domain);
 	fprintf(stderr, "\t\tmap = \"%s\"\n", nokey->map);
     }
@@ -782,11 +957,19 @@ ypresp_maplist *ypproc_maplist_2_svc(domainname *name,
 				     struct svc_req *rqstp)
 {
     static ypresp_maplist result;
+    struct sockaddr_in *rqhost;
     
 
+    rqhost = svc_getcaller(rqstp->rq_xprt);
+    if (!is_valid_host(rqhost))
+	return NULL;
+
     if (debug_flag)
     {
-	fprintf(stderr, "ypproc_maplist_2_svc():\n");
+	fprintf(stderr, "ypproc_maplist_2_svc(): [From: %s:%d]\n",
+		inet_ntoa(rqhost->sin_addr),
+		ntohs(rqhost->sin_port));
+		
 	fprintf(stderr, "\t\tdomain = \"%s\"\n", *name);
     }
 
@@ -825,7 +1008,7 @@ ypresp_maplist *ypproc_maplist_2_svc(domainname *name,
 	    closedir(dp);
 	}
     }
-
+    
     if (debug_flag)
     {
 	if (result.stat == YP_TRUE)
@@ -845,5 +1028,6 @@ ypresp_maplist *ypproc_maplist_2_svc(domainname *name,
 	else
 	    fprintf(stderr, "\t-> Error #%d\n", result.stat);
     }
+    
     return &result;
 }
